@@ -16,10 +16,10 @@ The main thread will intercept the following signals:
     - SIGHUP: reload configuration and log and changes done
     - SIGTERM: cleanly terminate all threads and exit
 
-A minimal communication protocol is implented, using the LISTEN/NOTIFY facility
-provided by postgres.  The dedicated main thread repository connection listens
-on the "powa_collector" channel.  A client, such as powa-web, can send requests
-on this channel and the main thread will act and respond accordingly.
+A minimal communication protocol is implemented, using the LISTEN/NOTIFY
+facility provided by postgres.  The dedicated main thread repository connection
+listens on the "powa_collector" channel.  A client, such as powa-web, can send
+requests on this channel and the main thread will act and respond accordingly.
 
 The requests are of the following form:
 
@@ -31,13 +31,19 @@ See the README.md file for the full protocol documentation.
 from powa_collector.options import (parse_options, get_full_config,
                                     add_servers_config)
 from powa_collector.powa_worker import PowaThread
+from powa_collector.customconn import get_connection
+from powa_collector.notify import (notify_parse_force_snapshot,
+                                   notify_parse_refresh_db_cat,
+                                   notify_allowed)
+from powa_collector.utils import conf_are_equal
 import psycopg2
 import select
 import logging
 import json
 import signal
+import time
 
-__VERSION__ = '1.2.0'
+__VERSION__ = '1.3.0'
 __VERSION_NUM__ = [int(part) for part in __VERSION__.split('.')]
 
 
@@ -69,6 +75,18 @@ class PowaCollector():
         signal.signal(signal.SIGHUP, self.sighandler)
         signal.signal(signal.SIGTERM, self.sighandler)
 
+
+    def __get_worker_by_srvid(self, srvid):
+        """Get the worker thread for the given srvid, if any"""
+        for k, worker in self.workers.items():
+            worker_srvid = self.config["servers"][k]["srvid"]
+
+            if (srvid != worker_srvid):
+                continue
+
+            return worker
+        return None
+
     def connect(self, options):
         """Connect to the repository
         Used for communication with powa-web and users of the communication repository
@@ -77,7 +95,9 @@ class PowaCollector():
         """
         try:
             self.logger.debug("Connecting on repository...")
-            self.__repo_conn = psycopg2.connect(options["repository"]['dsn'])
+            self.__repo_conn = get_connection(self.logger,
+                                              options["debug"],
+                                              options["repository"]['dsn'])
             self.__repo_conn.autocommit = True
             self.logger.debug("Connected.")
             cur = self.__repo_conn.cursor()
@@ -124,10 +144,10 @@ class PowaCollector():
 
         return True
 
-    def process_notification(self):
+    def process_notifications(self):
         """Process PostgreSQL NOTIFY messages.
         These come mainly from the UI, to ask us to reload our configuration,
-        or to display the workers status
+        or to display the workers status.
         """
         if (not self.__repo_conn):
             return
@@ -136,41 +156,42 @@ class PowaCollector():
         cur = self.__repo_conn.cursor()
 
         while (self.__repo_conn.notifies):
-            notif = self.__repo_conn.notifies.pop(0).payload.split(' ')
+            notify = self.__repo_conn.notifies.pop(0)
+            pid = notify.pid
+            payload = notify.payload.split(' ')
             status = ''
-            cmd = notif.pop(0)
+            cmd = payload.pop(0)
             channel = "-"
             status = "OK"
             data = None
 
             # the channel is mandatory, but if the message can be processed
             # without answering, we'll try to
-            if (len(notif) > 0):
-                channel = notif.pop(0)
+            if (len(payload) > 0):
+                channel = payload.pop(0)
 
             self.logger.debug("Received async command: %s %s %r" %
-                              (cmd, channel, notif))
+                              (cmd, channel, payload))
 
-            if (cmd == "RELOAD"):
-                self.reload_conf()
-                data = 'OK'
-            elif (cmd == "WORKERS_STATUS"):
-                # ignore the message if no channel was received
-                if (channel != '-'):
-                    # did the caller request a single server only?  We ignore
-                    # anything but the first parameter passed
-                    if (len(notif) > 0 and notif[0].isdigit()):
-                        w_id = int(notif[0])
-                        data = json.dumps(self.list_workers(w_id, False))
-                    else:
-                        data = json.dumps(self.list_workers(None, False))
-            # everything else is unhandled
+            if (not notify_allowed(pid, self.__repo_conn)):
+                status = 'ERROR'
+                data = 'Permission denied'
             else:
-                status = 'UNKNOWN'
-                data = ''
+                try:
+                    (status, data) = self.__process_one_notification(cmd,
+                                                                     channel,
+                                                                     payload)
+                except Exception as e:
+                    status = 'ERROR'
+                    data = str(e)
 
             # if there was a response channel, reply back
             if (channel != '-'):
+                # We need an extra {} around the data as the custom connection
+                # will call format() on the overall string, which would fail
+                # with a json dump as-is.  We therefore also need to make sure
+                # that any non-empty data looks like a json.
+
                 payload = ("%(cmd)s %(status)s %(data)s" %
                            {'cmd': cmd, 'status': status, 'data': data})
 
@@ -180,17 +201,75 @@ class PowaCollector():
                 # XXX we could implement multi-part answer, but if we ever
                 # reach that point, we should consider moving to a table
                 if (len(payload.encode('utf-8')) >= 8000):
-                    payload = ("%(cmd)s %(status)s %(data)" %
+                    payload = ("%(cmd)s %(status)s %(data)s" %
                                {'cmd': cmd,
                                 'status': "KO",
-                                'data': "ANSWER TOO LONG"})
+                                'data': "{ANSWER TOO LONG}"})
 
-                cur.execute("""NOTIFY "%(channel)s", '%(payload)s'""" %
-                            {'channel': channel,
-                             'payload': payload})
+                cur.execute('NOTIFY "' + channel + '", %(payload)s',
+                            {'payload': payload})
 
         cur.close()
 
+    def __process_one_notification(self, cmd, channel, notif):
+        """
+        Process a single notification, called by process_notifications.
+        """
+        status = "OK"
+        data = ''
+
+        if (cmd == "RELOAD"):
+            self.reload_conf()
+            data = 'OK'
+        elif (cmd == "WORKERS_STATUS"):
+            # ignore the message if no channel was received
+            if (channel != '-'):
+                # did the caller request a single server only?  We ignore
+                # anything but the first parameter passed
+                if (len(notif) > 0 and notif[0].isdigit()):
+                    w_id = int(notif[0])
+                    data = json.dumps(self.list_workers(w_id, False))
+                else:
+                    data = json.dumps(self.list_workers(None, False))
+        elif (cmd == "FORCE_SNAPSHOT"):
+            r_srvid = notify_parse_force_snapshot(notif)
+            worker = self.__get_worker_by_srvid(r_srvid)
+
+            if (worker is None):
+                raise Exception("Server id %d not found" % r_srvid)
+
+            request_time = time.time()
+            (status, data) = worker.request_forced_snapshot()
+
+            # If a snapshot could be scheduled, wait just a bit and see if we
+            # can report that the snapshot has already begun or will be started
+            # shortly.  It's possible, although unlikely, that the snapshot was
+            # started and finished during that time.  We don't try to properly
+            # detect that case, as it should only happen in toy setups.  We
+            # might want to change that later if we want to provide a way to
+            # inform callers of the completion of the snapshot they requested.
+            if (status == 'OK'):
+                time.sleep(0.1)
+                if worker.is_snapshot_in_progress():
+                    data = 'Snapshot in progress'
+                else:
+                    data = 'Snapshot wil begin shortly'
+        elif (cmd == "REFRESH_DB_CAT"):
+            (r_srvid, r_dbnames) = notify_parse_refresh_db_cat(notif)
+            worker = self.__get_worker_by_srvid(r_srvid)
+
+            if (worker is None):
+                raise Exception("Server id %d not found" % r_srvid)
+
+            (status, data) = worker.register_cat_refresh(r_dbnames)
+
+        # everything else is unhandled
+        else:
+            status = 'UNKNOWN'
+            data = 'Message "%s" invalid, command ignored' % cmd
+
+        return (status, data)
+
     def main(self):
         """Start the active loop.
         Connect or reconnect to the repository and starts threads to manage the
@@ -213,7 +292,8 @@ class PowaCollector():
             exit(1)
 
         for k, conf in self.config["servers"].items():
-            self.register_worker(k, self.config["repository"], conf)
+            self.register_worker(k, self.config["repository"], conf,
+                                 raw_options['debug'])
 
         self.list_workers()
 
@@ -233,15 +313,15 @@ class PowaCollector():
 
                 select.select([self.__repo_conn], [], [], 10)
 
-                self.process_notification()
+                self.process_notifications()
         except KeyboardInterrupt:
             self.logger.debug("KeyboardInterrupt caught")
             self.logger.info("Stopping all workers and exiting...")
             self.stop_all_workers()
 
-    def register_worker(self, name, repository, config):
+    def register_worker(self, name, repository, config, debug):
         """Add a worker thread to a server"""
-        self.workers[name] = PowaThread(name, repository, config)
+        self.workers[name] = PowaThread(name, repository, config, debug)
         self.workers[name].start()
 
     def stop_all_workers(self):
@@ -326,7 +406,8 @@ class PowaCollector():
             if (k not in self.workers or not self.workers[k].is_alive()):
                 self.logger.info("%s has been added, registering it..." % k)
                 self.register_worker(k, config_new["repository"],
-                                     config_new["servers"][k])
+                                     config_new["servers"][k],
+                                     config_new['debug'])
 
         # check for updated configuration
         for k in config_new["servers"]:
@@ -343,20 +424,3 @@ class PowaCollector():
 
         self.config = config_new
         self.logger.info('Reload done')
-
-
-def conf_are_equal(conf1, conf2):
-    """Compare two configurations, returns True if equal"""
-    for k in conf1.keys():
-        if (k not in conf2):
-            return False
-        if (conf1[k] != conf2[k]):
-            return False
-
-    for k in conf2.keys():
-        if (k not in conf1):
-            return False
-        if (conf1[k] != conf2[k]):
-            return False
-
-    return True
